# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
ResourceClient for Featureform.

DEPRECATED: This class is deprecated in favor of the unified Client class.
Use featureform.Client instead.

This module contains the legacy ResourceClient class for querying and listing
Featureform resources.
"""

import inspect
import logging
import os
import warnings
from typing import List, Optional, Union, get_args

from ..enums import ResourceType
from ..grpc_client import GrpcClient
from ..list import (
    list_name,
    list_name_status,
    list_name_variant_status,
    list_name_variant_status_desc,
)
from ..operations.cleanup import CleanupResult
from ..proto import metadata_pb2
from ..proto import metadata_pb2 as pb
from ..proto import metadata_pb2_grpc as ff_grpc
from ..resources import *
from ..status_display import StatusDisplayer, display_statuses
from ..tls import insecure_channel, secure_channel

__all__ = ["ResourceClient"]


class ResourceClient:
    """
    The resource client is used to retrieve information on specific resources
    (entities, providers, features, labels, training sets, models, users).

    Args:
        host (str): The hostname of the Featureform instance.
        insecure (bool): True if connecting to an insecure Featureform endpoint. False if using a self-signed or public TLS certificate
        cert_path (str): The path to a public certificate if using a self-signed certificate.

    **Using the Resource Client:**
    ``` py title="definitions.py"
    import featureform as ff
    from featureform import ResourceClient

    rc = ResourceClient("localhost:8000")

    # example query:
    redis = client.get_provider("redis-quickstart")
    ```
    """

    def __init__(
        self,
        host=None,
        local=False,
        insecure=False,
        cert_path=None,
        dry_run=False,
        debug=False,
    ):
        if local:
            raise Exception(
                "Local mode is not supported in this version. Use featureform <= 1.12.0 for localmode"
            )

        # This line ensures that the warning is only raised if ResourceClient is instantiated directly
        # TODO: Remove this check once ServingClient is deprecated
        is_instantiated_directed = inspect.stack()[1].function != "__init__"
        if is_instantiated_directed:
            warnings.warn(
                "ResourceClient is deprecated and will be removed in future versions; use Client instead.",
                PendingDeprecationWarning,
            )
        self._dry_run = dry_run
        self._stub = None
        self.local = local

        if dry_run:
            return

        host = host or os.getenv("FEATUREFORM_HOST")
        if host is None:
            raise RuntimeError(
                "If not in local mode then `host` must be passed or the environment"
                " variable FEATUREFORM_HOST must be set."
            )
        if insecure:
            channel = insecure_channel(host)
        else:
            channel = secure_channel(host, cert_path)
        self._stub = GrpcClient(ff_grpc.ApiStub(channel), debug=debug)
        self._host = host
        self._cert_path = cert_path or os.getenv("FEATUREFORM_CERT")
        self._insecure = insecure
        self.logger = logging.getLogger(__name__)

    def apply(
        self, asynchronous=False, verbose=False, cleanup=False
    ) -> Optional["CleanupResult"]:
        """
        Apply all definitions, creating and retrieving all specified resources.

        ```python
        import featureform as ff
        client = ff.Client()

        ff.register_postgres(
            host="localhost",
            port=5432,
        )

        client.apply()
        ```

        Args:
            asynchronous (bool): If True, apply will return immediately and not wait for resources to be created. If False, apply will wait for resources to be created and print out the status of each resource.

        """
        # Import here to avoid circular dependency
        from ..register import get_run, global_registrar, set_run

        try:
            resource_state = global_registrar.state()

            if resource_state.is_empty():
                print("No resources to apply")
                return

            print(f"Applying Run: {get_run()}")

            if self._dry_run:
                print(resource_state.sorted_list())
                return

            resource_state.create_all(
                self._stub,
                asynchronous,
                self._host,
                global_registrar.get_client_objects_for_resource(),
            )

            print()

            if not asynchronous and self._stub:
                resources = resource_state.sorted_list()
                display_statuses(self._stub, resources, self._host)

            if cleanup and not self._dry_run and self._stub:
                if asynchronous:
                    print(
                        "Warning: Cleanup is not supported in asynchronous mode. Skipping cleanup."
                    )
                return self._cleanup(resource_state, "total")
        finally:
            set_run("")
            global_registrar.clear_state()

    def delete(
        self,
        source: Union["DeletableResourceObjects", str],
        variant: Optional[str] = None,
        resource_type: Optional[ResourceType] = None,
        *,
        asynchronous: bool = True,
    ):
        """
        Delete a resource by name and variant or by resource object.

        There are three ways to delete a resource:
        1. Using a resource object (Feature, Label, Source, etc.)
        2. Using a provider object (OnlineProvider, OfflineProvider)
        3. Using a string name with required parameters

        Examples:
            ```python
            import featureform as ff
            client = ff.Client()

            # Delete using string name (requires variant and resource_type)
            client.delete("transactions", "kaggle", ff.ResourceType.SOURCE)

            # Delete using resource object
            feature = client.get_feature("my_feature", "v1")
            client.delete(feature)

            # Delete provider (no variant needed)
            client.delete("redis_provider", resource_type=ff.ResourceType.PROVIDER)
            ```

        Args:
            source: Either a resource object (Feature, Label, Source, etc.) or the name of the resource as a string.
                If a string is provided, resource_type is required and variant is required for non-provider resources.
            variant: Variant of the resource to delete. Required if source is a string and resource_type is not PROVIDER.
            resource_type: Type of resource to delete. Required if source is a string.
            asynchronous: If True, return immediately after issuing the delete request (default).

        Raises:
            ValueError: If source is a string and:
                - resource_type is not provided
                - variant is not provided (for non-provider resources)
            RuntimeError: If the resource does not exist
        """
        # Prepare the request based on the input type
        request = self._create_delete_request(source, variant, resource_type)

        # Send the request to delete the resource
        response = self._stub.MarkForDeletion(request)
        task_id = getattr(response, "task_id", "")

        resource_to_task = (
            (
                pb.ResourceType.Name(request.resource_id.resource_type),
                request.resource_id.resource.name,
                request.resource_id.resource.variant,
            ),
            task_id,
        )

        if asynchronous:
            print("Deleting resource async")
            return resource_to_task

        displayer = StatusDisplayer(self._stub, [])
        displayer.simple_poll({resource_to_task[0]: resource_to_task[1]})

        return resource_to_task

    def prune(
        self,
        source: Union["DeletableResourceObjects", str],
        variant: Optional[str] = None,
        resource_type: Optional[ResourceType] = None,
        *,
        asynchronous: bool = True,
    ):
        """Prune a resource and its unreferenced dependencies.

        Args:
            asynchronous: If True, return immediately after issuing the prune request (default).
        """
        if (
            isinstance(source, FeatureView)
            or resource_type == ResourceType.FEATURE_VIEW
        ):
            raise ValueError(
                "Pruning FeatureViews is not supported. Use `delete` instead."
            )
        # Import here to avoid circular dependency
        from ..providers import OfflineProvider, OnlineProvider

        request = self._create_prune_request(source, variant, resource_type)

        # Send the request to delete the resource
        response = self._stub.PruneResource(request)
        if (
            isinstance(source, str) and resource_type == ResourceType.PROVIDER
        ) or isinstance(source, (OfflineProvider, OnlineProvider)):
            print(
                "Run `delete` on provider after pruning to remove provider from Featureform"
            )
        resources_to_tasks = {
            (
                pb.ResourceType.Name(r.resource.resource_type),
                r.resource.resource.name,
                r.resource.resource.variant,
            ): r.task_id
            for r in response.results
        }

        if asynchronous:
            print("Pruning resource async")
            return resources_to_tasks

        if not resources_to_tasks:
            print("No resources to prune")
            return {}

        displayer = StatusDisplayer(self._stub, [])

        displayer.simple_poll(resources_to_tasks)

        return resources_to_tasks

    def get_task_statuses(self, task_ids: List[str]):
        """Fetch the latest status for the given task IDs."""
        request = metadata_pb2.GetTaskStatusesRequest(task_ids=task_ids)
        response = self._stub.GetTaskStatuses(request)
        return response.task_statuses

    def _print_detailed_cleanup_summary(
        self, asynchronous, stats, deleted_resources, failed_resources
    ):
        """
        Print a detailed summary of the cleanup operation.

        Args:
            asynchronous: Whether the cleanup was async
            stats: Dictionary with cleanup statistics
            deleted_resources: Set of successfully deleted resources
            failed_resources: List of resources that failed to delete
        """
        print(f"\n{'=' * 50}")

        if asynchronous:
            print("CLEANUP INITIATED (ASYNC)")
            print(f"{'=' * 50}")
            print(f"Total queued for deletion: {stats['queued']}")
            print(f"Total failed: {stats['failed']}")
        else:
            print("CLEANUP COMPLETE")
            print(f"{'=' * 50}")
            print(f"Total deleted: {stats['queued']}")
            print(f"Total failed: {stats['failed']}")

        # Print successfully deleted resources
        if deleted_resources:
            print(
                f"\n{'DELETED RESOURCES:' if not asynchronous else 'RESOURCES QUEUED FOR DELETION:'}"
            )
            print("-" * 30)
            # Group by resource type
            resources_by_type = {}
            for res_type, name, variant in deleted_resources:
                if res_type not in resources_by_type:
                    resources_by_type[res_type] = []
                resources_by_type[res_type].append((name, variant))

            for res_type, resources in resources_by_type.items():
                print(f"\n{res_type}:")
                for name, variant in sorted(resources):
                    if variant:
                        print(f"  - {name} (variant: {variant})")
                    else:
                        print(f"  - {name}")

        # Print failed resources
        if failed_resources:
            print(f"\nFAILED TO DELETE:")
            print("-" * 30)
            # Group by resource type
            failed_by_type = {}
            for res_type, name, variant in failed_resources:
                if res_type not in failed_by_type:
                    failed_by_type[res_type] = []
                failed_by_type[res_type].append((name, variant))

            for res_type, resources in failed_by_type.items():
                print(f"\n{res_type}:")
                for name, variant in sorted(resources):
                    if variant:
                        print(f"  - {name} (variant: {variant})")
                    else:
                        print(f"  - {name}")

        # Show detailed errors if any
        if stats["errors"] and self.logger.level <= logging.DEBUG:
            print("\nDETAILED ERRORS:")
            print("-" * 30)
            for resource_id, error in stats["errors"][:10]:
                print(f"  {resource_id}:")
                print(f"    Error: {error}")
            if len(stats["errors"]) > 10:
                print(f"  ... and {len(stats['errors']) - 10} more errors")

        # Final status message
        if not asynchronous:
            if stats["failed"] == 0:
                print(f"\n[SUCCESS] All resources cleaned up successfully")
            else:
                print(f"\n[WARNING] Cleanup completed with {stats['failed']} failures")
        else:
            print(f"\n[INFO] Resources will be deleted in the background")
            print(f"       Monitor progress in your Featureform dashboard")

        print(f"{'=' * 50}\n")

    def _cleanup(self, resource_state, mode, asynchronous=False) -> "CleanupResult":
        """
        Execute cleanup based on mode.

        Args:
            resource_state: Current state of resources that were just applied
            mode (str): Cleanup mode - currently only "total" is supported
            asynchronous (bool): If True, cleanup runs in background

        Returns:
            set: Resources that were deleted or queued for deletion
        """
        # Validate cleanup mode
        if mode != "total":
            self.logger.error(f"Only 'total' mode is currently supported, got '{mode}'")
            return set()

        print(f"\n{'=' * 50}")
        print(f"Starting cleanup (mode: {mode}, async: {asynchronous})")
        print(f"{'=' * 50}\n")

        # Collect resources to keep from current run
        resources_to_keep = self._get_resources_to_keep(resource_state)
        print(f"Resources to keep from current run: {len(resources_to_keep)}")
        for res_type, name, variant in resources_to_keep:
            print(f"  - {res_type}: {name} (variant: {variant})")

        # Find resources to delete
        resources_to_delete = self._find_resources_to_delete(resources_to_keep)

        if not resources_to_delete:
            return CleanupResult.empty()

        print(f"\nFound {len(resources_to_delete)} resources to delete")
        print("Resources to delete:")
        for res_type, name, variant in resources_to_delete:
            print(f"  - {res_type}: {name} (variant: {variant})")

        # Execute cleanup
        return self._execute_cleanup(resources_to_delete, asynchronous)

    def _get_resources_to_keep(self, resource_state):
        """
        Build set of resources from current run that should be kept.

        Args:
            resource_state: Current state of resources

        Returns:
            set: Tuples of (type, name, variant) to keep
        """
        resources_to_keep = set()

        for res in resource_state.sorted_list():
            if isinstance(res, ResourceVariant) or isinstance(res, FeatureView):
                resources_to_keep.add(res.to_key())

        return resources_to_keep

    def _find_resources_to_delete(self, resources_to_keep):
        """
        Find all resources that are not in the current run.

        Note: Feature views must be deleted first as they may depend on other resources.

        Args:
            resources_to_keep: Set of (type, name, variant) tuples to keep

        Returns:
            list: Tuples of (type, name, variant) to delete, with feature views first
        """
        # Define resource types and their list functions
        # IMPORTANT: Feature views MUST be first in this list
        resource_config = [
            (ResourceType.FEATURE_VIEW, self._stub.ListFeatureViews),
            (ResourceType.SOURCE_VARIANT, self._stub.ListSources),
            (ResourceType.FEATURE_VARIANT, self._stub.ListFeatures),
            (ResourceType.LABEL_VARIANT, self._stub.ListLabels),
            (ResourceType.TRAININGSET_VARIANT, self._stub.ListTrainingSets),
        ]

        resources_to_delete = []

        for resource_type, list_func in resource_config:
            try:
                resources = list_func(metadata_pb2.ListRequest())
                for item in resources:
                    # Feature views don't have variants, just use name directly
                    if resource_type == ResourceType.FEATURE_VIEW:
                        resource_tuple = (resource_type, item.name, "")
                        if resource_tuple not in resources_to_keep:
                            resources_to_delete.append(resource_tuple)
                    else:
                        # All other resource types have variants
                        for variant in getattr(item, "variants", []):
                            resource_tuple = (resource_type, item.name, variant)
                            if resource_tuple not in resources_to_keep:
                                resources_to_delete.append(resource_tuple)

            except Exception as e:
                self.logger.error(f"Failed to list {resource_type.name}: {e}")
                # Continue with other resource types

        return resources_to_delete

    def _execute_cleanup(self, resources_to_delete, asynchronous) -> "CleanupResult":
        """
        Execute the cleanup of resources.

        Args:
            resources_to_delete: List of (type, name, variant) tuples to delete
            asynchronous: If True, cleanup runs in background

        Returns:
            CleanupResult: Object containing cleanup results
        """
        deleted_resources = set()
        failed_resources = []
        stats = {"queued": 0, "failed": 0, "errors": []}

        operation = "Queuing" if asynchronous else "Deleting"
        print(f"\n{operation} resources...")

        for resource_type, name, variant in resources_to_delete:
            # Feature views don't have variants
            if resource_type == ResourceType.FEATURE_VIEW:
                resource_id = f"{resource_type} {name}"
            else:
                resource_id = f"{resource_type} {name}.{variant}"

            try:
                print(f"  {operation} {resource_id} ({resource_type.name})...")
                if (resource_type, name, variant) in deleted_resources:
                    print("    [SKIPPED] Already deleted or queued for deletion")
                    continue

                resources_to_tasks = {}
                # Feature views use delete, other resources use prune
                if resource_type == ResourceType.FEATURE_VIEW:
                    resource, task = self.delete(
                        name,
                        resource_type=resource_type,
                        asynchronous=asynchronous,
                    )
                    resources_to_tasks[resource] = task
                else:
                    resources_to_tasks = self.prune(
                        name,
                        variant,
                        resource_type=resource_type,
                        asynchronous=asynchronous,
                    )
                    resources_to_tasks = resources_to_tasks or {}

                if resources_to_tasks:
                    deleted_resources.update(resources_to_tasks.keys())
                    stats["queued"] += len(resources_to_tasks)
                    print(f"    [SUCCESS] {operation} completed for {resource_id}")

            except Exception as e:
                err_msg = str(e)
                if "Could not find resource to delete" in err_msg:
                    self.logger.debug(f"Skipping {resource_id}: {err_msg}")
                    continue

                stats["failed"] += 1
                stats["errors"].append((resource_id, str(e)))
                failed_resources.append((resource_type, name, variant))
                print(f"    [FAILED] Could not delete {resource_id}: {e}")

        # Print summary
        self._print_detailed_cleanup_summary(
            asynchronous, stats, deleted_resources, failed_resources
        )

        return CleanupResult(
            deleted_resources=deleted_resources,
            failed_resources=failed_resources,
            error_count=stats["failed"],
            errors=stats["errors"],
        )

    def _create_prune_request(
        self,
        source: Union["DeletableResourceObjects", str],
        variant: Optional[str] = None,
        resource_type: Optional[ResourceType] = None,
    ) -> metadata_pb2.PruneResourceRequest:
        return metadata_pb2.PruneResourceRequest(
            resource_id=self._build_resource_id(source, variant, resource_type)
        )

    def _create_delete_request(
        self,
        source: Union["DeletableResourceObjects", str],
        variant: Optional[str] = None,
        resource_type: Optional[ResourceType] = None,
    ) -> metadata_pb2.MarkForDeletionRequest:
        return metadata_pb2.MarkForDeletionRequest(
            resource_id=self._build_resource_id(source, variant, resource_type)
        )

    def _build_resource_id(
        self,
        source: Union["DeletableResourceObjects", str],
        variant: Optional[str] = None,
        resource_type: Optional[ResourceType] = None,
    ) -> metadata_pb2.ResourceID:
        """Helper to construct a ResourceID for prune or delete requests."""
        # Import here to avoid circular dependency
        from ..providers import OfflineProvider, OnlineProvider
        from ..register import DeletableResourceObjects

        if isinstance(source, str):
            if resource_type is None:
                raise ValueError(
                    "resource_type must be specified if source is a string"
                )

            if not ResourceType.is_deletable(resource_type):
                raise ValueError("resource_type must be deletable")

            # handle resources without variants
            if resource_type == ResourceType.PROVIDER or resource_type == FeatureView:
                return metadata_pb2.ResourceID(
                    resource=metadata_pb2.NameVariant(name=source),
                    resource_type=resource_type.to_proto(),
                )

            if not variant and resource_type != ResourceType.FEATURE_VIEW:
                raise ValueError("variant must be specified for non-provider resources")

            return metadata_pb2.ResourceID(
                resource=metadata_pb2.NameVariant(name=source, variant=variant),
                resource_type=resource_type.to_proto(),
            )

        if isinstance(source, FeatureView):
            return metadata_pb2.ResourceID(
                resource=metadata_pb2.NameVariant(name=source.name),
                resource_type=ResourceType.FEATURE_VIEW.to_proto(),
            )

        if isinstance(source, (OfflineProvider, OnlineProvider)):
            return metadata_pb2.ResourceID(
                resource=metadata_pb2.NameVariant(name=source.name()),
                resource_type=ResourceType.PROVIDER.to_proto(),
            )

        if not isinstance(source, get_args(DeletableResourceObjects)):
            raise ValueError("resource is not deletable")

        name, variant = source.name_variant()
        return metadata_pb2.ResourceID(
            resource=metadata_pb2.NameVariant(name=name, variant=variant),
            resource_type=source.get_resource_type().to_proto(),
        )

    def run(self):
        """
        Run tasks for all definitions, creating and retrieving all specified resources.

        ```python
        import featureform as ff
        client = ff.Client()

        ff.register_postgres(
            host="localhost",
            port=5432,
        )

        client.run()
        ```
        """
        # Import here to avoid circular dependency
        from ..register import global_registrar

        try:
            resource_state = global_registrar.state()
            if resource_state.is_empty():
                print("No resources to run")
                return

            if self._dry_run:
                print(resource_state.sorted_list())
                return

            resource_state.run_all(
                self._stub, global_registrar.get_client_objects_for_resource()
            )

        finally:
            global_registrar.clear_state()

    def get_auth_config(self):
        if not self.local:
            return self._stub.GetAuthConfig(metadata_pb2.Empty())
        return None

    def get_user(self, name, local=False):
        """Get a user. Prints out name of user, and all resources associated with the user.

        **Examples:**

        ``` py title="Input"
        featureformer = rc.get_user("featureformer")
        ```

        ``` json title="Output"
        // get_user prints out formatted information on user
        USER NAME:                     featureformer
        -----------------------------------------------

        NAME                           VARIANT                        TYPE
        avg_transactions               quickstart                     feature
        fraudulent                     quickstart                     label
        fraud_training                 quickstart                     training set
        transactions                   kaggle                         source
        average_user_transaction       quickstart                     source
        -----------------------------------------------
        ```

        ``` py title="Input"
        print(featureformer)
        ```

        ``` json title="Output"
        // get_user returns the User object

        name: "featureformer"
        features {
        name: "avg_transactions"
        variant: "quickstart"
        }
        labels {
        name: "fraudulent"
        variant: "quickstart"
        }
        trainingsets {
        name: "fraud_training"
        variant: "quickstart"
        }
        sources {
        name: "transactions"
        variant: "kaggle"
        }
        sources {
        name: "average_user_transaction"
        variant: "quickstart"
        }
        ```

        Args:
            name (str): Name of user to be retrieved

        Returns:
            user (User): User
        """
        return get_user_info(self._stub, name)

    def get_entity(self, name, local=False):
        """Get an entity. Prints out information on entity, and all resources associated with the entity.

        **Examples:**

        ``` py title="Input"
        entity = rc.get_entity("user")
        ```

        ``` json title="Output"
        // get_entity prints out formatted information on entity

        ENTITY NAME:                   user
        STATUS:                        NO_STATUS
        -----------------------------------------------

        NAME                           VARIANT                        TYPE
        avg_transactions               quickstart                     feature
        fraudulent                     quickstart                     label
        fraud_training                 quickstart                     training set
        -----------------------------------------------
        ```

        ``` py title="Input"
        print(postgres)
        ```

        ``` json title="Output"
        // get_entity returns the Entity object

        name: "user"
        features {
            name: "avg_transactions"
            variant: "quickstart"
        }
        labels {
            name: "fraudulent"
            variant: "quickstart"
        }
        trainingsets {
            name: "fraud_training"
            variant: "quickstart"
        }
        ```
        """
        return get_entity_info(self._stub, name)

    def get_model(self, name, local=False) -> Model:
        """Get a model. Prints out information on model, and all resources associated with the model.

        Args:
            name (str): Name of model to be retrieved

        Returns:
            model (Model): Model
        """
        model = None
        model_proto = get_resource_info(self._stub, "model", name)
        if model_proto is not None:
            model = Model(model_proto.name, description="", tags=[], properties={})

        return model

    def get_provider(self, name, local=False):
        """Get a provider. Prints out information on provider, and all resources associated with the provider.

        **Examples:**

        ``` py title="Input"
        postgres = client.get_provider("postgres-quickstart")
        ```

        ``` json title="Output"
        // get_provider prints out formatted information on provider

        NAME:                          postgres-quickstart
        DESCRIPTION:                   A Postgres deployment we created for the Featureform quickstart
        TYPE:                          POSTGRES_OFFLINE
        SOFTWARE:                      postgres
        STATUS:                        NO_STATUS
        -----------------------------------------------
        SOURCES:
        NAME                           VARIANT
        transactions                   kaggle
        average_user_transaction       quickstart
        -----------------------------------------------
        FEATURES:
        NAME                           VARIANT
        -----------------------------------------------
        LABELS:
        NAME                           VARIANT
        fraudulent                     quickstart
        -----------------------------------------------
        TRAINING SETS:
        NAME                           VARIANT
        fraud_training                 quickstart
        -----------------------------------------------
        ```

        ``` py title="Input"
        print(postgres)
        ```

        ``` json title="Output"
        // get_provider returns the Provider object

        name: "postgres-quickstart"
        description: "A Postgres deployment we created for the Featureform quickstart"
        type: "POSTGRES_OFFLINE"
        software: "postgres"
        serialized_config: "{\"Host\": \"quickstart-postgres\",
                            \"Port\": \"5432\",
                            \"Username\": \"postgres\",
                            \"Password\": \"password\",
                            \"Database\": \"postgres\"}"
        sources {
        name: "transactions"
        variant: "kaggle"
        }
        sources {
        name: "average_user_transaction"
        variant: "quickstart"
        }
        trainingsets {
        name: "fraud_training"
        variant: "quickstart"
        }
        labels {
        name: "fraudulent"
        variant: "quickstart"
        }
        ```

        Args:
            name (str): Name of provider to be retrieved

        Returns:
            provider (Provider): Provider
        """
        return get_provider_info(self._stub, name)

    def get_feature(self, name, variant):
        return FeatureVariant.get_by_name_variant(self._stub, name, variant)

    def print_feature(self, name, variant=None, local=False):
        """Get a feature. Prints out information on feature, and all variants associated with the feature. If variant is included, print information on that specific variant and all resources associated with it.

        **Examples:**

        ``` py title="Input"
        avg_transactions = rc.get_feature("avg_transactions")
        ```

        ``` json title="Output"
        // get_feature prints out formatted information on feature

        NAME:                          avg_transactions
        STATUS:                        NO_STATUS
        -----------------------------------------------
        VARIANTS:
        quickstart                     default
        -----------------------------------------------
        ```

        ``` py title="Input"
        print(avg_transactions)
        ```

        ``` json title="Output"
        // get_feature returns the Feature object

        name: "avg_transactions"
        default_variant: "quickstart"
        variants: "quickstart"
        ```

        ``` py title="Input"
        avg_transactions_variant = ff.get_feature("avg_transactions", "quickstart")
        ```

        ``` json title="Output"
        // get_feature with variant provided prints out formatted information on feature variant

        NAME:                          avg_transactions
        VARIANT:                       quickstart
        TYPE:                          float32
        ENTITY:                        user
        OWNER:                         featureformer
        PROVIDER:                      redis-quickstart
        STATUS:                        NO_STATUS
        -----------------------------------------------
        SOURCE:
        NAME                           VARIANT
        average_user_transaction       quickstart
        -----------------------------------------------
        TRAINING SETS:
        NAME                           VARIANT
        fraud_training                 quickstart
        -----------------------------------------------
        ```

        ``` py title="Input"
        print(avg_transactions_variant)
        ```

        ``` json title="Output"
        // get_feature returns the FeatureVariant object

        name: "avg_transactions"
        variant: "quickstart"
        source {
        name: "average_user_transaction"
        variant: "quickstart"
        }
        type: "float32"
        entity: "user"
        created {
        seconds: 1658168552
        nanos: 142461900
        }
        owner: "featureformer"
        provider: "redis-quickstart"
        trainingsets {
        name: "fraud_training"
        variant: "quickstart"
        }
        columns {
        entity: "user_id"
        value: "avg_transaction_amt"
        }
        ```

        Args:
            name (str): Name of feature to be retrieved
            variant (str): Name of variant of feature

        Returns:
            feature (Union[Feature, FeatureVariant]): Feature or FeatureVariant
        """
        if not variant:
            return get_resource_info(self._stub, "feature", name)
        return get_feature_variant_info(self._stub, name, variant)

    def _get_all_and_latest_variants(
        self, resource_name, resource_type, get_latest_variant=False
    ):
        # Import here to avoid circular dependency
        from ..register import Variants
        from ..registrar import (
            ColumnSourceRegistrar,
            FeatureColumnResource,
            LabelColumnResource,
            SubscriptableTransformation,
        )

        if isinstance(resource_name, str):
            res_name = resource_name
            if resource_type is None:
                raise ValueError(
                    "A resource type param must be provided if the resource name param is of type string."
                )
            res_type = resource_type
        elif isinstance(
            resource_name,
            (
                FeatureColumnResource,
                LabelColumnResource,
                TrainingSetVariant,
                ColumnSourceRegistrar,
                SubscriptableTransformation,
            ),
        ):
            res_name = resource_name.name_variant()[0]
            res_type = resource_name.get_resource_type()
        elif isinstance(resource_name, Variants):
            if resource_type is None:
                raise ValueError(
                    f"{resource_name} is of type Variants. Please provide a resource type."
                )
            for _, res in resource_name.resources.items():
                res_name = res.name
                break
            res_type = resource_type
        else:
            raise ValueError(
                f"Expected input: string, Feature, Label, Source or Training Set object, actual input: {resource_name}."
            )

        search_name = metadata_pb2.NameRequest(name=metadata_pb2.Name(name=res_name))

        stub_get_functions = {
            ResourceType.FEATURE_VARIANT: self._stub.GetFeatures,
            ResourceType.ONDEMAND_FEATURE: self._stub.GetFeatures,
            ResourceType.LABEL_VARIANT: self._stub.GetLabels,
            ResourceType.SOURCE_VARIANT: self._stub.GetSources,
            ResourceType.TRAININGSET_VARIANT: self._stub.GetTrainingSets,
        }

        try:
            get_func = stub_get_functions[res_type]
        except KeyError:
            raise ValueError(
                f"Resource type {res_type.to_string()} doesnt have variants."
            )
        # Only the variants of the first resource are needed
        for x in get_func(iter([search_name])):
            return x.default_variant if get_latest_variant else x.variants

    def get_variants(self, resource_name, resource_type=None):
        """
        Get all variants of a resource.

        Args:
            resource_name (Union[str, FeatureColumnResource, LabelColumnResource, TrainingSetVariant, ColumnSourceRegistrar, Variants]): Name of resource or Resource object
            resource_type (ResourceType): Type of resource

        Returns:
            variants (List[str]): List of variants of the resource
        """
        return self._get_all_and_latest_variants(
            resource_name, resource_type, get_latest_variant=False
        )

    def latest_variant(self, resource_name, resource_type=None):
        """
        Get the most recent variant of a resource.

        Args:
            resource_name (Union[str, FeatureColumnResource, LabelColumnResource, TrainingSetVariant, ColumnSourceRegistrar, Variants]): Name of resource or Resource object
            resource_type (ResourceType): Type of resource

        Returns:
            variants (str): Latest variant of the resource
        """
        return self._get_all_and_latest_variants(
            resource_name, resource_type, get_latest_variant=True
        )

    def get_label(self, name, variant):
        return LabelVariant.get_by_name_variant(self._stub, name, variant)

    def print_label(self, name, variant=None, local=False):
        """Get a label. Prints out information on label, and all variants associated with the label. If variant is included, print information on that specific variant and all resources associated with it.

        **Examples:**

        ``` py title="Input"
        fraudulent = rc.get_label("fraudulent")
        ```

        ``` json title="Output"
        // get_label prints out formatted information on label

        NAME:                          fraudulent
        STATUS:                        NO_STATUS
        -----------------------------------------------
        VARIANTS:
        quickstart                     default
        -----------------------------------------------
        ```

        ``` py title="Input"
        print(fraudulent)
        ```

        ``` json title="Output"
        // get_label returns the Label object

        name: "fraudulent"
        default_variant: "quickstart"
        variants: "quickstart"
        ```

        ``` py title="Input"
        fraudulent_variant = ff.get_label("fraudulent", "quickstart")
        ```

        ``` json title="Output"
        // get_label with variant provided prints out formatted information on label variant

        NAME:                          fraudulent
        VARIANT:                       quickstart
        TYPE:                          bool
        ENTITY:                        user
        OWNER:                         featureformer
        PROVIDER:                      postgres-quickstart
        STATUS:                        NO_STATUS
        -----------------------------------------------
        SOURCE:
        NAME                           VARIANT
        transactions                   kaggle
        -----------------------------------------------
        TRAINING SETS:
        NAME                           VARIANT
        fraud_training                 quickstart
        -----------------------------------------------
        ```

        ``` py title="Input"
        print(fraudulent_variant)
        ```

        ``` json title="Output"
        // get_label returns the LabelVariant object

        name: "fraudulent"
        variant: "quickstart"
        type: "bool"
        source {
        name: "transactions"
        variant: "kaggle"
        }
        entity: "user"
        created {
        seconds: 1658168552
        nanos: 154924300
        }
        owner: "featureformer"
        provider: "postgres-quickstart"
        trainingsets {
        name: "fraud_training"
        variant: "quickstart"
        }
        columns {
        entity: "customerid"
        value: "isfraud"
        }
        ```

        Args:
            name (str): Name of label to be retrieved
            variant (str): Name of variant of label

        Returns:
            label (Union[label, LabelVariant]): Label or LabelVariant
        """
        if not variant:
            return get_resource_info(self._stub, "label", name)
        return get_label_variant_info(self._stub, name, variant)

    def get_training_set(self, name, variant):
        return TrainingSetVariant.get_by_name_variant(self._stub, name, variant)

    def print_training_set(self, name, variant=None, local=False):
        """Get a training set. Prints out information on training set, and all variants associated with the training set. If variant is included, print information on that specific variant and all resources associated with it.

        **Examples:**

        ``` py title="Input"
        fraud_training = rc.get_training_set("fraud_training")
        ```

        ``` json title="Output"
        // get_training_set prints out formatted information on training set

        NAME:                          fraud_training
        STATUS:                        NO_STATUS
        -----------------------------------------------
        VARIANTS:
        quickstart                     default
        -----------------------------------------------
        ```

        ``` py title="Input"
        print(fraud_training)
        ```

        ``` json title="Output"
        // get_training_set returns the TrainingSet object

        name: "fraud_training"
        default_variant: "quickstart"
        variants: "quickstart"
        ```

        ``` py title="Input"
        fraudulent_variant = ff.get_training set("fraudulent", "quickstart")
        ```

        ``` json title="Output"
        // get_training_set with variant provided prints out formatted information on training set variant

        NAME:                          fraud_training
        VARIANT:                       quickstart
        OWNER:                         featureformer
        PROVIDER:                      postgres-quickstart
        STATUS:                        NO_STATUS
        -----------------------------------------------
        LABEL:
        NAME                           VARIANT
        fraudulent                     quickstart
        -----------------------------------------------
        FEATURES:
        NAME                           VARIANT
        avg_transactions               quickstart
        -----------------------------------------------
        ```

        ``` py title="Input"
        print(fraudulent_variant)
        ```

        ``` json title="Output"
        // get_training_set returns the TrainingSetVariant object

        name: "fraud_training"
        variant: "quickstart"
        owner: "featureformer"
        created {
        seconds: 1658168552
        nanos: 157934800
        }
        provider: "postgres-quickstart"
        features {
        name: "avg_transactions"
        variant: "quickstart"
        }
        label {
        name: "fraudulent"
        variant: "quickstart"
        }
        ```

        Args:
            name (str): Name of training set to be retrieved
            variant (str): Name of variant of training set

        Returns:
            training_set (Union[TrainingSet, TrainingSetVariant]): TrainingSet or TrainingSetVariant
        """
        if not variant:
            return get_resource_info(self._stub, "training-set", name)
        return get_training_set_variant_info(self._stub, name, variant)

    def get_source(self, name, variant):
        # Import here to avoid circular dependency
        from ..register import global_registrar
        from ..registrar import ColumnSourceRegistrar

        source_variant = SourceVariant.get_by_name_variant(self._stub, name, variant)
        return ColumnSourceRegistrar(global_registrar, source_variant)

    def print_source(self, name, variant=None, local=False):
        """Get a source. Prints out information on source, and all variants associated with the source. If variant is included, print information on that specific variant and all resources associated with it.

        **Examples:**

        ``` py title="Input"
        transactions = rc.get_transactions("transactions")
        ```

        ``` json title="Output"
        // get_source prints out formatted information on source

        NAME:                          transactions
        STATUS:                        NO_STATUS
        -----------------------------------------------
        VARIANTS:
        kaggle                         default
        -----------------------------------------------
        ```

        ``` py title="Input"
        print(transactions)
        ```

        ``` json title="Output"
        // get_source returns the Source object

        name: "transactions"
        default_variant: "kaggle"
        variants: "kaggle"
        ```

        ``` py title="Input"
        transactions_variant = rc.get_source("transactions", "kaggle")
        ```

        ``` json title="Output"
        // get_source with variant provided prints out formatted information on source variant

        NAME:                          transactions
        VARIANT:                       kaggle
        OWNER:                         featureformer
        DESCRIPTION:                   Fraud Dataset From Kaggle
        PROVIDER:                      postgres-quickstart
        STATUS:                        NO_STATUS
        -----------------------------------------------
        DEFINITION:
        TRANSFORMATION

        -----------------------------------------------
        SOURCES
        NAME                           VARIANT
        -----------------------------------------------
        PRIMARY DATA
        Transactions
        FEATURES:
        NAME                           VARIANT
        -----------------------------------------------
        LABELS:
        NAME                           VARIANT
        fraudulent                     quickstart
        -----------------------------------------------
        TRAINING SETS:
        NAME                           VARIANT
        fraud_training                 quickstart
        -----------------------------------------------
        ```

        ``` py title="Input"
        print(transactions_variant)
        ```

        ``` json title="Output"
        // get_source returns the SourceVariant object

        name: "transactions"
        variant: "kaggle"
        owner: "featureformer"
        description: "Fraud Dataset From Kaggle"
        provider: "postgres-quickstart"
        created {
        seconds: 1658168552
        nanos: 128768000
        }
        trainingsets {
        name: "fraud_training"
        variant: "quickstart"
        }
        labels {
        name: "fraudulent"
        variant: "quickstart"
        }
        primaryData {
        table {
            name: "Transactions"
        }
        }
        ```

        Args:
            name (str): Name of source to be retrieved
            variant (str): Name of variant of source

        Returns:
            source (Union[Source, SourceVariant]): Source or SourceVariant
        """
        if not variant:
            return get_resource_info(self._stub, "source", name)
        return get_source_variant_info(self._stub, name, variant)

    def list_features(self, local=False):
        """List all features.

        **Examples:**
        ``` py title="Input"
        features_list = rc.list_features()
        ```

        ``` json title="Output"
        // list_features prints out formatted information on all features

        NAME                           VARIANT                        STATUS
        user_age                       quickstart (default)           READY
        avg_transactions               quickstart (default)           READY
        avg_transactions               production                     CREATED
        ```

        ``` py title="Input"
        print(features_list)
        ```

        ``` json title="Output"
        // list_features returns a list of Feature objects

        [name: "user_age"
        default_variant: "quickstart"
        variants: "quickstart"
        , name: "avg_transactions"
        default_variant: "quickstart"
        variants: "quickstart"
        variants: "production"
        ]
        ```

        Returns:
            features (List[Feature]): List of Feature Objects
        """
        return list_name_variant_status(self._stub, "feature")

    def list_labels(self, local=False):
        """List all labels.

        **Examples:**
        ``` py title="Input"
        features_list = rc.list_labels()
        ```

        ``` json title="Output"
        // list_labels prints out formatted information on all labels

        NAME                           VARIANT                        STATUS
        user_age                       quickstart (default)           READY
        avg_transactions               quickstart (default)           READY
        avg_transactions               production                     CREATED
        ```

        ``` py title="Input"
        print(label_list)
        ```

        ``` json title="Output"
        // list_features returns a list of Feature objects

        [name: "user_age"
        default_variant: "quickstart"
        variants: "quickstart"
        , name: "avg_transactions"
        default_variant: "quickstart"
        variants: "quickstart"
        variants: "production"
        ]
        ```

        Returns:
            labels (List[Label]): List of Label Objects
        """
        return list_name_variant_status(self._stub, "label")

    def list_users(self, local=False):
        """List all users. Prints a list of all users.

        **Examples:**
        ``` py title="Input"
        users_list = rc.list_users()
        ```

        ``` json title="Output"
        // list_users prints out formatted information on all users

        NAME                           STATUS
        featureformer                  NO_STATUS
        featureformers_friend          CREATED
        ```

        ``` py title="Input"
        print(features_list)
        ```

        ``` json title="Output"
        // list_features returns a list of Feature objects

        [name: "featureformer"
        features {
        name: "avg_transactions"
        variant: "quickstart"
        }
        labels {
        name: "fraudulent"
        variant: "quickstart"
        }
        trainingsets {
        name: "fraud_training"
        variant: "quickstart"
        }
        sources {
        name: "transactions"
        variant: "kaggle"
        }
        sources {
        name: "average_user_transaction"
        variant: "quickstart"
        },
        name: "featureformers_friend"
        features {
        name: "user_age"
        variant: "production"
        }
        sources {
        name: "user_profiles"
        variant: "production"
        }
        ]
        ```

        Returns:
            users (List[User]): List of User Objects
        """
        return list_name_status(self._stub, "user")

    def list_entities(self, local=False):
        """List all entities. Prints a list of all entities.

        **Examples:**
        ``` py title="Input"
        entities = rc.list_entities()
        ```

        ``` json title="Output"
        // list_entities prints out formatted information on all entities

        NAME                           STATUS
        user                           CREATED
        transaction                    CREATED
        ```

        ``` py title="Input"
        print(features_list)
        ```

        ``` json title="Output"
        // list_entities returns a list of Entity objects

        [name: "user"
        features {
        name: "avg_transactions"
        variant: "quickstart"
        }
        features {
        name: "avg_transactions"
        variant: "production"
        }
        features {
        name: "user_age"
        variant: "quickstart"
        }
        labels {
        name: "fraudulent"
        variant: "quickstart"
        }
        trainingsets {
        name: "fraud_training"
        variant: "quickstart"
        }
        ,
        name: "transaction"
        features {
        name: "amount_spent"
        variant: "production"
        }
        ]
        ```

        Returns:
            entities (List[Entity]): List of Entity Objects
        """
        return list_name_status(self._stub, "entity")

    def list_sources(self, local=False):
        """List all sources. Prints a list of all sources.

        **Examples:**
        ``` py title="Input"
        sources_list = rc.list_sources()
        ```

        ``` json title="Output"
        // list_sources prints out formatted information on all sources

        NAME                           VARIANT                        STATUS                         DESCRIPTION
        average_user_transaction       quickstart (default)           NO_STATUS                      the average transaction amount for a user
        transactions                   kaggle (default)               NO_STATUS                      Fraud Dataset From Kaggle
        ```

        ``` py title="Input"
        print(sources_list)
        ```

        ``` json title="Output"
        // list_sources returns a list of Source objects

        [name: "average_user_transaction"
        default_variant: "quickstart"
        variants: "quickstart"
        , name: "transactions"
        default_variant: "kaggle"
        variants: "kaggle"
        ]
        ```

        Returns:
            sources (List[Source]): List of Source Objects
        """
        return list_name_variant_status_desc(self._stub, "source")

    def list_training_sets(self, local=False):
        """List all training sets. Prints a list of all training sets.

        **Examples:**
        ``` py title="Input"
        training_sets_list = rc.list_training_sets()
        ```

        ``` json title="Output"
        // list_training_sets prints out formatted information on all training sets

        NAME                           VARIANT                        STATUS                         DESCRIPTION
        fraud_training                 quickstart (default)           READY                          Training set for fraud detection.
        fraud_training                 v2                             CREATED                        Improved training set for fraud detection.
        recommender                    v1 (default)                   CREATED                        Training set for recommender system.
        ```

        ``` py title="Input"
        print(training_sets_list)
        ```

        ``` json title="Output"
        // list_training_sets returns a list of TrainingSet objects

        [name: "fraud_training"
        default_variant: "quickstart"
        variants: "quickstart", "v2",
        name: "recommender"
        default_variant: "v1"
        variants: "v1"
        ]
        ```

        Returns:
            training_sets (List[TrainingSet]): List of TrainingSet Objects
        """
        return list_name_variant_status_desc(self._stub, "training-set")

    def list_models(self, local=False) -> List[Model]:
        """List all models. Prints a list of all models.

        Returns:
            models (List[Model]): List of Model Objects
        """
        model_protos = list_name(self._stub, "model")
        # TODO: apply values from proto
        models = [Model(proto.name, tags=[], properties={}) for proto in model_protos]

        return models

    def list_providers(self, local=False):
        """List all providers. Prints a list of all providers.

        **Examples:**
        ``` py title="Input"
        providers_list = rc.list_providers()
        ```

        ``` json title="Output"
        // list_providers prints out formatted information on all providers

        NAME                           STATUS                         DESCRIPTION
        redis-quickstart               CREATED                      A Redis deployment we created for the Featureform quickstart
        postgres-quickstart            CREATED                      A Postgres deployment we created for the Featureform quickst
        ```

        ``` py title="Input"
        print(providers_list)
        ```

        ``` json title="Output"
        // list_providers returns a list of Providers objects

        [name: "redis-quickstart"
        description: "A Redis deployment we created for the Featureform quickstart"
        type: "REDIS_ONLINE"
        software: "redis"
        serialized_config: "{\"Addr\": \"quickstart-redis:6379\", \"Password\": \"\", \"DB\": 0}"
        features {
        name: "avg_transactions"
        variant: "quickstart"
        }
        features {
        name: "avg_transactions"
        variant: "production"
        }
        features {
        name: "user_age"
        variant: "quickstart"
        }
        , name: "postgres-quickstart"
        description: "A Postgres deployment we created for the Featureform quickstart"
        type: "POSTGRES_OFFLINE"
        software: "postgres"
        serialized_config: "{\"Host\": \"quickstart-postgres\", \"Port\": \"5432\", \"Username\": \"postgres\", \"Password\": \"password\", \"Database\": \"postgres\"}"
        sources {
        name: "transactions"
        variant: "kaggle"
        }
        sources {
        name: "average_user_transaction"
        variant: "quickstart"
        }
        trainingsets {
        name: "fraud_training"
        variant: "quickstart"
        }
        labels {
        name: "fraudulent"
        variant: "quickstart"
        }
        ]
        ```

        Returns:
            providers (List[Provider]): List of Provider Objects
        """
        return list_name_status_desc(self._stub, "provider")

    def search(self, raw_query, local=False):
        """Search for registered resources. Prints a list of results.

        **Examples:**
        ``` py title="Input"
        providers_list = rc.search("transact")
        ```

        ``` json title="Output"
        // search prints out formatted information on all matches

        NAME                           VARIANT            TYPE
        avg_transactions               default            Source
        ```
        """
        if type(raw_query) != str or len(raw_query) == 0:
            raise Exception("query must be string and cannot be empty")
        processed_query = raw_query.translate({ord(i): None for i in ".,-@!*#"})
        return search(processed_query, self._host)

    def list_feature_views(self, local=False):
        """List all feature views. Prints a list of all feature views."""
        return list_name_status(self._stub, "feature_view")

    def get_feature_view(self, name, local=False):
        """Get a feature view. Prints out information on feature view, and all resources associated with the feature view."""
        return get_feature_view_info(self._stub, name)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Exceptions - Re-exports from core.exceptions for backwards compatibility.

DEPRECATED: Import from featureform.core.exceptions instead.
"""

# Re-export from core.exceptions for backwards compatibility
from .core.exceptions import (
    FeatureformException,
    FeatureNotFound,
    InvalidSQLQuery,
    InvalidTrainingSetFeatureComputationMode,
    LabelNotFound,
)

__all__ = [
    "FeatureformException",
    "FeatureNotFound",
    "LabelNotFound",
    "InvalidSQLQuery",
    "InvalidTrainingSetFeatureComputationMode",
]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Credential classes for various cloud providers and services.
"""

import json
import os
import re
import sys
from dataclasses import dataclass, field
from typing import Union

from typeguard import typechecked


def read_file(file_path: str) -> str:
    """Read file content as string."""
    with open(file_path, "r") as f:
        return f.read()


@typechecked
@dataclass
class AWSStaticCredentials:
    """
    Static Credentials for an AWS services.

    **Example**
    ```
    aws_credentials = ff.AWSStaticCredentials(
        access_key="<AWS_ACCESS_KEY>",
        secret_key="<AWS_SECRET_KEY>"
    )
    ```

    Args:
        access_key (str): AWS Access Key.
        secret_key (str): AWS Secret Key.
    """

    access_key: str = field(default="")
    secret_key: str = field(default="")
    _type: str = field(default="AWS_STATIC_CREDENTIALS")

    def __post_init__(
        self,
    ):
        if self.access_key == "":
            raise Exception("'AWSStaticCredentials' access_key cannot be empty")

        if self.secret_key == "":
            raise Exception("'AWSStaticCredentials' secret_key cannot be empty")

    @staticmethod
    def type() -> str:
        return "AWS_STATIC_CREDENTIALS"

    def config(self):
        return {
            "AccessKeyId": self.access_key,
            "SecretKey": self.secret_key,
            "Type": self._type,
        }


@typechecked
@dataclass
class AWSAssumeRoleCredentials:
    """
    Assume Role Credentials for an AWS services.

    If an IAM role for service account (IRSA) has been configured, the default credentials provider chain
        will be used to get the credentials stored on the pod. See the following link for more information:
        https://docs.aws.amazon.com/eks/latest/userguide/iam-roles-for-service-accounts.html

    **Example**
    ```
    aws_credentials = ff.AWSAssumeRoleCredentials()
    ```
    """

    _type: str = field(default="AWS_ASSUME_ROLE_CREDENTIALS")

    @staticmethod
    def type() -> str:
        return "AWS_ASSUME_ROLE_CREDENTIALS"

    def config(self):
        return {
            "Type": self._type,
        }


@typechecked
@dataclass
class GCPCredentials:
    def __init__(
        self, project_id: str, credentials_path: str = "", credential_json: dict = None
    ):
        """
        Credentials for an GCP.

        **Example**
        ```
        gcp_credentials = ff.GCPCredentials(
            project_id="<project_id>",
            credentials_path="<path_to_credentials>"
        )
        ```

        Args:
            project_id (str): The project id.
            credentials_path (str): The path to the credentials file.
        """
        if project_id == "":
            raise Exception("'GCPCredentials' project_id cannot be empty")

        self.project_id = project_id
        self.credentials_path = credentials_path

        if self.credentials_path == "" and credential_json is None:
            raise ValueError(
                "Either a path to credentials or json credentials must be provided"
            )
        elif self.credentials_path != "" and credential_json is not None:
            raise ValueError(
                "Only one of credentials_path or credentials_json can be provided"
            )
        elif self.credentials_path != "" and credential_json is None:
            if not os.path.isfile(credentials_path):
                raise Exception(
                    f"'GCPCredentials' credentials_path '{credentials_path}' file not found"
                )
            with open(credentials_path) as f:
                self.credentials = json.load(f)
        else:
            self.credentials = credential_json

    def type(self):
        return "GCPCredentials"

    def config(self):
        return {
            "ProjectId": self.project_id,
            "JSON": self.credentials,
        }

    def to_json(self):
        return self.credentials


@typechecked
@dataclass
class BasicCredentials:
    def __init__(
        self,
        username: str,
        password: str = "",
    ):
        """
        Credentials for an EMR cluster.

        **Example**
        ```
        creds = ff.BasicCredentials(
            username="<username>",
            password="<password>",
        )

        hdfs = ff.register_hdfs(
            name="hdfs",
            credentials=creds,
            ...
        )
        ```

        Args:
            username (str): the username of account.
            password (str): the password of account (optional).
        """
        self.username = username
        self.password = password

    def type(self):
        return "BasicCredential"

    def config(self):
        return {
            "Username": self.username,
            "Password": self.password,
        }


@typechecked
@dataclass
class KerberosCredentials:
    def __init__(
        self,
        username: str,
        password: str,
        krb5s_file: str,
    ):
        """
        Credentials for an EMR cluster.

        **Example**
        ```
        kerberos = ff.KerberosCredentials(
            username="<username>",
            password="<password>",
            krb5s_file="<path/to/krb5s_file>",
        )

        hdfs = ff.register_hdfs(
            name="hdfs",
            credentials=kerberos,
            ...
        )
        ```

        Args:
            username (str): the username of account.
            password (str): the password of account.
            krb5s_file (str): the path to krb5s file.
        """
        self.username = username
        self.password = password
        self.krb5s_conf = read_file(krb5s_file)

    def type(self):
        return "KerberosCredential"

    def config(self):
        return {
            "Username": self.username,
            "Password": self.password,
            "Krb5Conf": self.krb5s_conf,
        }


# Constants for Pyspark Versions
MAJOR_VERSION = "3"
MINOR_VERSIONS = ["9", "10", "11", "12"]


@typechecked
@dataclass
class DatabricksCredentials:
    """
    Credentials for a Databricks cluster.

    **Example**
    ```
    databricks = ff.DatabricksCredentials(
        username="<my_username>",
        password="<my_password>",
        host="<databricks_hostname>",
        token="<databricks_token>",
        cluster_id="<databricks_cluster>",
    )

    spark = ff.register_spark(
        name="spark",
        executor=databricks,
        ...
    )
    ```

    Args:
        username (str): Username for a Databricks cluster.
        password (str): Password for a Databricks cluster.
        host (str): The hostname of a Databricks cluster.
        token (str): The token for a Databricks cluster.
        cluster_id (str): ID of an existing Databricks cluster.
    """

    username: str = ""
    password: str = ""
    host: str = ""
    token: str = ""
    cluster_id: str = ""

    def __post_init__(self):
        host_token_provided = (
            self.username == ""
            and self.password == ""
            and self.host != ""
            and self.token != ""
        )
        username_password_provided = (
            self.username != ""
            and self.password != ""
            and self.host == ""
            and self.token == ""
        )

        if (
            not host_token_provided
            and not username_password_provided
            or host_token_provided
            and username_password_provided
        ):
            raise Exception(
                "The DatabricksCredentials requires only one credentials set ('username' and 'password' or 'host' and 'token' set.)"
            )

        if not self.cluster_id:
            raise Exception("Cluster_id of existing cluster must be provided")

        if not self._validate_cluster_id():
            raise ValueError(
                f"Invalid cluster_id: expected id in the format 'xxxx-xxxxxx-xxxxxxxx' but received '{self.cluster_id}'"
            )

        if self.host and not self._validate_token():
            raise ValueError(
                f"Invalid token: expected token in the format 'dapi' + 32 alphanumeric characters (optionally ending with '-' and 1 alphanumeric character) but received '{self.token}'"
            )

    def _validate_cluster_id(self):
        cluster_id_regex = r"^\w{4}-\w{6}-\w{8}$"
        return re.match(cluster_id_regex, self.cluster_id)

    def _validate_token(self):
        token_regex = r"^dapi[a-zA-Z0-9]{32}(-[a-zA-Z0-9])?$"
        return re.match(token_regex, self.token)

    def type(self):
        return "DATABRICKS"

    def config(self):
        return {
            "Username": self.username,
            "Password": self.password,
            "Host": self.host,
            "Token": self.token,
            "Cluster": self.cluster_id,
        }


@typechecked
@dataclass
class EMRCredentials:
    def __init__(
        self,
        emr_cluster_id: str,
        emr_cluster_region: str,
        credentials: Union[AWSStaticCredentials, AWSAssumeRoleCredentials],
    ):
        """
        Credentials for an EMR cluster.

        **Example**
        ```
        emr = ff.EMRCredentials(
            emr_cluster_id="<cluster_id>",
            emr_cluster_region="<cluster_region>",
            credentials="<AWS_Credentials>",
        )

        spark = ff.register_spark(
            name="spark",
            executor=emr,
            ...
        )
        ```

        Args:
            emr_cluster_id (str): ID of an existing EMR cluster.
            emr_cluster_region (str): Region of an existing EMR cluster.
            credentials (Union[AWSStaticCredentials, AWSAssumeRoleCredentials]): Credentials for an AWS account with access to the cluster
        """
        self.emr_cluster_id = emr_cluster_id
        self.emr_cluster_region = emr_cluster_region
        self.credentials = credentials

    def type(self):
        return "EMR"

    def config(self):
        return {
            "ClusterName": self.emr_cluster_id,
            "ClusterRegion": self.emr_cluster_region,
            "Credentials": self.credentials.config(),
        }


@typechecked
@dataclass
class SparkCredentials:
    def __init__(
        self,
        master: str,
        deploy_mode: str,
        python_version: str,
        core_site_path: str = "",
        yarn_site_path: str = "",
    ):
        """
        Credentials for a Generic Spark Cluster

        **Example**
        ```
        spark_credentials = ff.SparkCredentials(
            master="yarn",
            deploy_mode="cluster",
            python_version="3.7.12",
            core_site_path="core-site.xml",
            yarn_site_path="yarn-site.xml"
        )

        spark = ff.register_spark(
            name="spark",
            executor=spark_credentials,
            ...
        )
        ```

        Args:
            master (str): The hostname of the Spark cluster. (The same that would be passed to `spark-submit`).
            deploy_mode (str): The deploy mode of the Spark cluster. (The same that would be passed to `spark-submit`).
            python_version (str): The Python version running on the cluster. Supports 3.9-3.12
            core_site_path (str): The path to the core-site.xml file. (For Yarn clusters only)
            yarn_site_path (str): The path to the yarn-site.xml file. (For Yarn clusters only)
        """
        self.master = master.lower()
        self.deploy_mode = deploy_mode.lower()
        self.core_site_path = core_site_path
        self.yarn_site_path = yarn_site_path

        if self.deploy_mode != "cluster" and self.deploy_mode != "client":
            raise Exception(
                f"Spark does not support '{self.deploy_mode}' deploy mode. It only supports 'cluster' and 'client'."
            )

        self.python_version = self._verify_python_version(
            self.deploy_mode, python_version
        )

        self._verify_yarn_config()

    def _verify_python_version(self, deploy_mode, version):
        if deploy_mode == "cluster" and version == "":
            client_python_version = sys.version_info
            client_major = str(client_python_version.major)
            client_minor = str(client_python_version.minor)

            if client_major != MAJOR_VERSION:
                client_major = "3"
            if client_minor not in MINOR_VERSIONS:
                client_minor = "7"

            version = f"{client_major}.{client_minor}"

        if version.count(".") == 2:
            major, minor, _ = version.split(".")
        elif version.count(".") == 1:
            major, minor = version.split(".")
        else:
            raise Exception(
                "Please specify your Python version on the Spark cluster. Accepted formats: Major.Minor or Major.Minor.Patch; ex. '3.9' or '3.9.16"
            )

        if major != MAJOR_VERSION or minor not in MINOR_VERSIONS:
            raise Exception(
                f"The Python version {version} is not supported. Currently, supported versions are 3.9-3.12."
            )

        """
        The Python versions on the Docker image are 3.9.16, 3.10.10, and 3.11.2.
        This conditional statement sets the patch number based on the minor version.
        """
        if minor == "10":
            patch = "10"
        elif minor == "11":
            patch = "2"
        else:
            patch = "16"

        return f"{major}.{minor}.{patch}"

    def _verify_yarn_config(self):
        if self.master == "yarn" and (
            self.core_site_path == "" or self.yarn_site_path == ""
        ):
            raise Exception(
                "Yarn requires core-site.xml and yarn-site.xml files."
                "Please copy these files from your Spark instance to local, then provide the local path in "
                "core_site_path and yarn_site_path. "
            )

    def type(self):
        return "SPARK"

    def config(self):
        core_site = (
            "" if self.core_site_path == "" else open(self.core_site_path, "r").read()
        )
        yarn_site = (
            "" if self.yarn_site_path == "" else open(self.yarn_site_path, "r").read()
        )

        return {
            "Master": self.master,
            "DeployMode": self.deploy_mode,
            "PythonVersion": self.python_version,
            "CoreSite": core_site,
            "YarnSite": yarn_site,
        }


# Type alias for executor credentials
ExecutorCredentials = Union[DatabricksCredentials, EMRCredentials, SparkCredentials]



from __future__ import annotations

import os
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, Generic, TypeVar, Union


class SecretType(str, Enum):
    ENVIRONMENT = "environment"
    STATIC = "static"
    VAULT = "vault"


S = TypeVar("S", bound="Secret")


class SecretProvider(Generic[S], ABC):
    @abstractmethod
    def secret(self, key: str, **kwargs) -> S:
        """Create a secret retriever for delayed access."""
        pass


@dataclass(frozen=True)
class EnvironmentSecretProvider(SecretProvider["EnvironmentSecret"]):
    """Creates references to environment variable secrets."""

    def secret(self, key: str, **kwargs) -> EnvironmentSecret:
        """Create an environment variable secret reference."""
        return EnvironmentSecret(key)


@dataclass(frozen=True)
class VaultSecretProvider(SecretProvider["VaultSecret"]):
    """Creates references to Vault secrets."""

    default_path: str
    vault_provider_name: str

    def secret(self, key: str, **kwargs) -> VaultSecret:
        """
        Create a Vault secret reference.

        Args:
            key: Secret key
            **kwargs: Optional parameters including:
                path: Optional override for default_path
        """
        path = kwargs.get("path", self.default_path)
        return VaultSecret(key, path, self.vault_provider_name)


class Secret(ABC):
    """Abstract base class for secret values with delayed retrieval."""

    @abstractmethod
    def get(self) -> Any:
        """Retrieve the secret value."""
        pass

    @abstractmethod
    def serialize(self) -> Dict[str, Any]:
        """Serialize the secret configuration."""
        pass

    @classmethod
    def deserialize(cls, data: Union[Dict[str, Any] | Any]) -> Secret:
        """
        Deserialize a secret from a dictionary configuration.

        Args:
            data: Dictionary containing serialized secret data

        Returns:
            Appropriate Secret implementation

        Raises:
            ValueError: If the secret type is unknown
        """
        # Handle static values (backwards compatibility)
        if "type" not in data:
            return StaticSecret(data)

        secret_map = {
            SecretType.ENVIRONMENT.name: EnvironmentSecret,
            SecretType.STATIC.name: StaticSecret,
            SecretType.VAULT.name: VaultSecret,
        }

        secret_type = data["type"]
        if secret_type.upper() not in secret_map:
            raise ValueError(f"Unknown secret type: {secret_type}")

        return secret_map[secret_type].deserialize(data)


@dataclass(frozen=True)
class StaticSecret(Secret):
    """Secret containing a static value."""

    value: Any

    def get(self) -> Any:
        return self.value

    def serialize(self) -> str:
        return self.value

    @classmethod
    def deserialize(cls, data: Dict[str, Any] | str) -> StaticSecret:
        if not isinstance(data, Dict):
            return cls(data)  # Backwards compatibility
        return cls(data.get("value", data))


@dataclass(frozen=True)
class EnvironmentSecret(Secret):
    """Secret that retrieves its value from environment variables."""

    key: str

    def get(self) -> str | None:
        return os.getenv(self.key)

    def serialize(self) -> Dict[str, str]:
        return {"type": SecretType.ENVIRONMENT.name, "key": self.key}

    @classmethod
    def deserialize(cls, data: Dict[str, Any]) -> EnvironmentSecret:
        if "key" not in data:
            raise ValueError(
                "EnvironmentSecret deserialization failed: 'key' is missing."
            )
        return cls(data["key"])


@dataclass(frozen=True)
class VaultSecret(Secret):
    """Secret that retrieves its value from Vault."""

    key: str
    path: str
    _vault_provider_name: str

    def get(self) -> str | None:
        raise NotImplementedError("VaultSecret retrieval not implemented")

    def serialize(self) -> Dict[str, str]:
        return {
            "type": SecretType.VAULT.name,
            "key": self.key,
            "path": self.path,
            "vault_provider_name": self._vault_provider_name,
        }

    @classmethod
    def deserialize(cls, data: Dict[str, Any]) -> VaultSecret:
        for field in ["key", "path", "vault_provider_name"]:
            if field not in data:
                raise ValueError(
                    f"VaultSecret deserialization failed: '{field}' is missing."
                )

        return cls(
            data["key"],
            data["path"],
            data["vault_provider_name"],
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
State management - Re-exports from core.state for backwards compatibility.

DEPRECATED: Import from featureform.core.state instead.
"""

# Re-export from core.state for backwards compatibility
from ..core.state import ResourceRedefinedError, ResourceState

__all__ = [
    "ResourceState",
    "ResourceRedefinedError",
]



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Input validation utilities for transformations.

This module provides utilities to detect and validate input types for
streaming and batch transformations.
"""

from typing import Any, List

from ..resources.locations import StreamingInput

__all__ = [
    "is_streaming_input",
    "validate_streaming_transformation_inputs",
    "validate_batch_transformation_inputs",
]


def is_streaming_input(input_source: Any) -> bool:
    """
    Determine if an input source is a streaming input.

    Streaming inputs are sources that extend the StreamingInput base class,
    such as KafkaTopic.

    Args:
        input_source: The input source to check.

    Returns:
        bool: True if the input is a streaming input, False otherwise.
    """
    return isinstance(input_source, StreamingInput)


def validate_streaming_transformation_inputs(inputs: List[Any]) -> List[Any]:
    """
    Validate inputs for a streaming transformation.

    Streaming transformations require all inputs to be streaming inputs.

    Args:
        inputs: List of input sources for the transformation.

    Returns:
        The validated list of streaming inputs.

    Raises:
        ValueError: If no inputs are provided or if any non-streaming inputs are found.
    """
    if not inputs:
        raise ValueError(
            "Streaming transformations require at least one input. "
            "No inputs were provided."
        )

    non_streaming_inputs = [inp for inp in inputs if not is_streaming_input(inp)]

    if non_streaming_inputs:
        non_streaming_count = len(non_streaming_inputs)
        raise ValueError(
            f"Streaming transformations can only have streaming inputs. "
            f"Found {non_streaming_count} non-streaming input(s). "
            f"All inputs must be streaming inputs."
        )

    return inputs


def validate_batch_transformation_inputs(inputs: List[Any]) -> List[Any]:
    """
    Validate inputs for a batch transformation.

    Batch transformations should not have streaming inputs.

    Args:
        inputs: List of input sources for the transformation.

    Returns:
        The validated list of inputs.

    Raises:
        ValueError: If any streaming inputs are found.
    """
    if not inputs:
        return inputs

    streaming_inputs = [inp for inp in inputs if is_streaming_input(inp)]

    if streaming_inputs:
        streaming_count = len(streaming_inputs)
        raise ValueError(
            f"Batch transformations cannot have streaming inputs. "
            f"Found {streaming_count} streaming input(s). "
            f"Use streaming_sql_transformation or streaming_df_transformation instead."
        )

    return inputs



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Entity resource classes.
"""

from dataclasses import dataclass, field
from typing import Tuple, Union

import grpc
from typeguard import typechecked

from ..enums import OperationType, ResourceType
from ..proto import metadata_pb2 as pb
from .provider import Properties


@typechecked
@dataclass
class Entity:
    """Entity resource representing a business object."""

    name: str
    description: str
    status: str = "NO_STATUS"
    tags: list = field(default_factory=list)
    properties: dict = field(default_factory=dict)

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.CREATE

    @staticmethod
    def get_resource_type() -> ResourceType:
        return ResourceType.ENTITY

    def _create(self, req_id, stub) -> Tuple[None, None]:
        serialized = pb.EntityRequest(
            entity=pb.Entity(
                name=self.name,
                description=self.description,
                tags=pb.Tags(tag=self.tags),
                properties=Properties(self.properties).serialized,
            ),
            request_id="",
        )

        stub.CreateEntity(serialized)
        return None, None

    def to_dictionary(self):
        return {
            "name": self.name,
            "description": self.description,
            "status": self.status,
            "tags": self.tags,
            "properties": self.properties,
        }


class EntityReference:
    """Reference to an entity resource."""

    name: str
    obj: Union[Entity, None]

    @staticmethod
    def operation_type() -> OperationType:
        return OperationType.GET

    @staticmethod
    def get_resource_type() -> ResourceType:
        return ResourceType.ENTITY

    def _get(self, stub):
        entityList = stub.GetEntities(
            iter([pb.NameRequest(name=pb.Name(name=self.name))])
        )
        try:
            for entity in entityList:
                self.obj = entity
        except grpc._channel._MultiThreadedRendezvous:
            raise ValueError(f"Entity {self.name} not found.")



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Mapping classes for Featureform resources.

This module contains classes for mapping entities and columns in features and labels,
including ResourceColumnMapping, EntityMapping, EntityMappings, and schema classes.
"""

from dataclasses import dataclass
from typing import List, Optional

from typeguard import typechecked

from ..enums import ScalarType
from ..proto import metadata_pb2 as pb


@typechecked
@dataclass
class ResourceColumnMapping:
    """Maps entity, value, and timestamp columns for a resource."""

    entity: str
    value: str
    timestamp: str = ""

    def proto(self) -> pb.Columns:
        return pb.Columns(
            entity=self.entity,
            value=self.value,
            ts=self.timestamp,
        )

    def to_entity_mappings_proto(self, entity_name: str) -> pb.EntityMappings:
        mapping = pb.EntityMapping(name=entity_name, entity_column=self.entity)
        return pb.EntityMappings(
            mappings=[mapping],
            value_column=self.value,
            timestamp_column=self.timestamp,
        )


@typechecked
@dataclass
class ColumnSchema:
    """Schema definition for a column."""

    name: str
    value_type: ScalarType

    def to_proto(self) -> pb.ColumnSchema:
        return pb.ColumnSchema(name=self.name, type=self.value_type.to_proto())

    @staticmethod
    def from_proto(proto: pb.ColumnSchema) -> "ColumnSchema":
        if proto is None:
            raise ValueError("Feature definitions require column schemas")
        if proto.type is None or proto.type.WhichOneof("Type") != "scalar":
            raise ValueError("Feature definitions require scalar column schemas")
        return ColumnSchema(
            name=proto.name,
            value_type=ScalarType.from_proto(proto.type.scalar),
        )


@typechecked
@dataclass
class FeatureColumn:
    """Schema for a feature column with optional timestamp."""

    feature_col: ColumnSchema
    timestamp_col: Optional[ColumnSchema] = None

    @staticmethod
    def from_proto(proto: pb.FeatureColumn) -> "FeatureColumn":
        return FeatureColumn(
            feature_col=ColumnSchema(
                name=proto.feature_col.name,
                value_type=ScalarType.from_proto(proto.feature_col.type.scalar),
            ),
            timestamp_col=(
                ColumnSchema(
                    name=proto.timestamp_col.name,
                    value_type=ScalarType.from_proto(proto.timestamp_col.type.scalar),
                )
                if proto.HasField("timestamp_col")
                else None
            ),
        )


@typechecked
@dataclass
class FeaturesSchema:
    """Schema for features including entity and feature columns."""

    entity_col: ColumnSchema
    feature_cols: list[FeatureColumn]

    @staticmethod
    def from_proto(proto: pb.FeaturesSchema) -> "FeaturesSchema":
        entity_col = ColumnSchema(
            name=proto.entity_col.name,
            value_type=ScalarType.from_proto(proto.entity_col.type.scalar),
        )
        feature_cols = [
            FeatureColumn.from_proto(feature_col) for feature_col in proto.feature_cols
        ]
        return FeaturesSchema(entity_col, feature_cols)


@typechecked
@dataclass
class EntityMapping:
    """Mapping between entity names and columns."""

    name: str
    entity_column: str

    def to_proto(self):
        return pb.EntityMapping(
            name=self.name,
            entity_column=self.entity_column,
        )


@typechecked
@dataclass
class EntityMappings:
    """Collection of entity mappings with value and timestamp columns."""

    mappings: List[EntityMapping]
    value_column: str
    timestamp_column: Optional[str] = None

    def to_proto(self):
        return pb.EntityMappings(
            mappings=[m.to_proto() for m in self.mappings],
            value_column=self.value_column,
            timestamp_column=self.timestamp_column,
        )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

"""
Column resource classes for Featureform.

This module contains classes for feature, label, and embedding column resources.
"""

from abc import ABC
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

from ..config.offline_stores import ResourceSnowflakeConfig
from ..enums import ResourceType, ScalarType
from ..utils.helpers import set_tags_properties

if TYPE_CHECKING:
    from ..resources import Entity, NameVariant
    from . import UserRegistrar
    from .column_mapping import ColumnMapping


__all__ = [
    "ColumnResource",
    "FeatureColumnResource",
    "LabelColumnResource",
    "EmbeddingColumnResource",
]


class ColumnResource(ABC):
    """
    Base class for all column resources. This class is not meant to be instantiated directly.
    In the original syntax, features and labels were registered using the `register_resources`
    method on the sources (e.g. SQL/DF transformation or tables sources); however, in the new
    Class API syntax, features and labels can now be declared as class attributes on an entity
    class. This means that all possible params for either resource must be passed into this base
    class prior to calling `register_column_resources` on the registrar.
    """

    def __init__(
        self,
        transformation_args: tuple,
        type: Union[ScalarType, str],
        resource_type: ResourceType,
        entity: Union["Entity", str],
        owner: Union[str, "UserRegistrar"],
        timestamp_column: str,
        description: str,
        schedule: str,
        tags: List[str],
        properties: Dict[str, str],
        inference_store: Union[str, "OnlineProvider", "FileStoreProvider"] = "",
        name: str = "",
        variant: str = "",
        resource_snowflake_config: Optional[ResourceSnowflakeConfig] = None,
    ):
        registrar, source_name_variant, columns = transformation_args
        self.type = type if isinstance(type, str) else type.value
        self.registrar = registrar
        self.source = source_name_variant
        self.entity_column = columns[0]
        self.source_column = columns[1]
        self.resource_type = resource_type
        self.entity = entity
        self.name = name
        self.owner = owner
        self.inference_store = inference_store
        if not timestamp_column and len(columns) == 3:
            self.timestamp_column = columns[2]
        elif timestamp_column and len(columns) == 3:
            raise Exception("Timestamp column specified twice.")
        else:
            self.timestamp_column = timestamp_column
        self.description = description
        self.schedule = schedule
        tags, properties = set_tags_properties(tags, properties)
        self.tags = tags
        self.properties = properties
        self.variant = variant
        self.resource_snowflake_config = resource_snowflake_config

    def register(self):
        features, labels = self.get_resources_by_type(self.resource_type)

        self.registrar.register_column_resources(
            source=self.source,
            entity=self.entity,
            entity_column=self.entity_column,
            owner=self.owner,
            inference_store=self.inference_store,
            features=features,
            labels=labels,
            timestamp_column=self.timestamp_column,
            schedule=self.schedule,
            client_object=self,
        )

    def get_resources_by_type(
        self, resource_type: ResourceType
    ) -> Tuple[List["ColumnMapping"], List["ColumnMapping"]]:
        resources = [
            {
                "name": self.name,
                "variant": self.variant,
                "column": self.source_column,
                "type": self.type,
                "description": self.description,
                "tags": self.tags,
                "properties": self.properties,
                "resource_snowflake_config": self.resource_snowflake_config,
            }
        ]

        if resource_type == ResourceType.FEATURE_VARIANT:
            features = resources
            labels = []
        elif resource_type == ResourceType.LABEL_VARIANT:
            features = []
            labels = resources
        else:
            raise ValueError(
                f"Resource type {self.resource_type.to_string()} not supported"
            )
        return (features, labels)

    def name_variant(self) -> "NameVariant":
        if self.name is None:
            raise ValueError("Resource name not set")
        if self.variant is None:
            raise ValueError("Resource variant not set")
        return (self.name, self.variant)

    def get_resource_type(self) -> ResourceType:
        return self.resource_type

    def to_key(self) -> tuple[ResourceType, str, str]:
        return self.resource_type, self.name, self.variant


# Variants class moved to registrar/variants.py


class FeatureColumnResource(ColumnResource):
    def __init__(
        self,
        transformation_args: tuple,
        type: Union[ScalarType, str],
        entity: Union["Entity", str] = "",
        name: str = "",
        variant: str = "",
        owner: str = "",
        inference_store: Union[str, "OnlineProvider", "FileStoreProvider"] = "",
        timestamp_column: str = "",
        description: str = "",
        schedule: str = "",
        tags: Optional[List[str]] = None,
        properties: Optional[Dict[str, str]] = None,
        resource_snowflake_config: Optional[ResourceSnowflakeConfig] = None,
    ):
        """
        Feature registration object.

        **Example**
        ```
        @ff.entity
        class Customer:
        # Register a column from a transformation as a feature
            transaction_amount = ff.Feature(
                fare_per_family_member[["CustomerID", "Amount", "Transaction Time"]],
                variant="quickstart",
                type=ff.Float64,
                inference_store=redis,
            )
        ```

        Args:
            transformation_args (tuple): A transformation or source function and the columns name in the format: <transformation_function>[[<entity_column>, <value_column>, <timestamp_column (optional)>]].
            variant (str): An optional variant name for the feature.
            type (Union[ScalarType, str]): The type of the value in for the feature.
            inference_store (Union[str, "OnlineProvider", "FileStoreProvider"]): Where to store for online serving.
        """
        super().__init__(
            transformation_args=transformation_args,
            type=type,
            resource_type=ResourceType.FEATURE_VARIANT,
            entity=entity,
            name=name,
            variant=variant,
            owner=owner,
            inference_store=inference_store,
            timestamp_column=timestamp_column,
            description=description,
            schedule=schedule,
            tags=tags,
            properties=properties,
            resource_snowflake_config=resource_snowflake_config,
        )


class LabelColumnResource(ColumnResource):
    def __init__(
        self,
        transformation_args: tuple,
        type: Union[ScalarType, str],
        entity: Union["Entity", str] = "",
        name: str = "",
        variant: str = "",
        owner: str = "",
        timestamp_column: str = "",
        description: str = "",
        schedule: str = "",
        tags: List[str] = [],
        properties: Dict[str, str] = {},
    ):
        """
        Label registration object.

        **Example**
        ```
        @ff.entity
        class Customer:
        # Register a column from a transformation as a label
            transaction_amount = ff.Label(
                fare_per_family_member[["CustomerID", "Amount", "Transaction Time"]],
                variant="quickstart",
                type=ff.Float64
            )
        ```

        Args:
            transformation_args (tuple): A transformation or source function and the columns name in the format: <transformation_function>[[<entity_column>, <value_column>, <timestamp_column (optional)>]]
            variant (str): An optional variant name for the label.
            type (Union[ScalarType, str]): The type of the value in for the label.
        """
        super().__init__(
            transformation_args=transformation_args,
            type=type,
            resource_type=ResourceType.LABEL_VARIANT,
            entity=entity,
            name=name,
            variant=variant,
            owner=owner,
            timestamp_column=timestamp_column,
            description=description,
            schedule=schedule,
            tags=tags,
            properties=properties,
        )


class EmbeddingColumnResource(ColumnResource):
    def __init__(
        self,
        transformation_args: tuple,
        dims: int,
        vector_db: Union[str, "OnlineProvider", "FileStoreProvider"],
        entity: Union["Entity", str] = "",
        name="",
        variant="",
        owner: str = "",
        timestamp_column: str = "",
        description: str = "",
        schedule: str = "",
        tags: List[str] = [],
        properties: Dict[str, str] = {},
    ):
        """
        Embedding Feature registration object.

        **Example**
        ```
        @ff.entity
        class Speaker:
        # Register a column from a transformation as a label
            transaction_amount = ff.Embedding(
                vectorize_comments[["PK", "Vector"]],
                dims=384,
                vector_db=pinecone,
                description="Embeddings created from speakers' comments in episodes",
                variant="v1"
            )
        ```

        Args:
            transformation_args (tuple): A transformation or source function and the columns name in the format: <transformation_function>[[<entity_column>, <value_column>]]
            dims (int): Dimensionality of the embedding.
            vector_db (Union[str, OnlineProvider]): The name of the vector database to store the embeddings in.
            variant (str): An optional variant name for the feature.
            description (str): An optional description for the feature.
        """
        super().__init__(
            transformation_args=transformation_args,
            type=ScalarType.FLOAT32,
            resource_type=ResourceType.FEATURE_VARIANT,
            entity=entity,
            name=name,
            variant=variant,
            owner=owner,
            inference_store=vector_db,
            timestamp_column=timestamp_column,
            description=description,
            schedule=schedule,
            tags=tags,
            properties=properties,
        )
        if dims < 1:
            raise ValueError("Vector dimensions must be a positive integer")
        self.dims = dims

    def get_resources_by_type(
        self, resource_type: ResourceType
    ) -> Tuple[List["ColumnMapping"], List["ColumnMapping"]]:
        features, labels = super().get_resources_by_type(resource_type)
        features[0]["dims"] = self.dims
        features[0]["is_embedding"] = True
        return (features, labels)
